/* Dim a tone with PWM
 *
 * We dim a tone of 400 Hz with a period of 1 second on a buzzer.  We use a
 * sample rate of 48 kHz and oversample with a factor of 20.  These 20 other
 * samples allow us to vary the level (the volume of the tone) by adjusting the
 * duty cycle.  A duty cycle of about 90% will yield a high volume and vice
 * versa.
 *
 * This code is a bit complicated because we are handling two periods that are
 * independent.  First, we have a period that determines the frequency of the
 * tone that we hear.  Second, we have a period with which we dim.  These two
 * periods are independent of each other but they have to be "driven" from the
 * same function.
 */

#include "pico/stdlib.h"
#include <stdio.h>
#include <math.h>
#include "hardware/pwm.h"
#include "hardware/clocks.h"
#include "../xiao_rp2040.h"

#define PIN_BUZZER D8

#define FREQ_CLOCK_SYS_BUS 125000000

#define SAMPLE_RATE 48000
#define OVERSAMPLING_FACTOR 20

// for the value of wrap we are aiming for about 128 such that we can set a
// level between 0 and 128.  This particular value is about 130.
#define WRAP ((uint16_t) (FREQ_CLOCK_SYS_BUS / (SAMPLE_RATE * OVERSAMPLING_FACTOR)))

#define TIME_WRAP (1.0 / FREQ_CLOCK_SYS_BUS * WRAP)

/* We aim for a wrap of 130 which means that a high volume and low volume
 * for the sound is what is defined below
 */
#define HIGH_LEVEL 128
#define LOW_LEVEL 16

// the tone that we want to produce
#define NOTE_HZ 400

// one second defined in terms of the number of wraps, which is about 960000.
#define ONE_SECOND (FREQ_CLOCK_SYS_BUS / WRAP)

/* The number of wraps to determine a new level
 *
 * Every time the number of wraps is this value, we adapt the level to a new
 * value between the high level and the low level.  The factor 2 is because we
 * want the level to rise and fall.  This value is a function of the number of
 * wraps and the range between the high and low level.
 */
#define NR_WRAPS_DET_LEVEL (ONE_SECOND / ((HIGH_LEVEL - LOW_LEVEL) * 2))


// a wrap will occur 960000 times per second
// to get a tone of 400 Hz, it should turn on and off every 48000 * 20 / 400
void on_pwm_wrap() {
  // In this function we will count two independent number of wraps with
  // different periods.  The first number of wraps determines the level with
  // which we want to pulse.  In essence, it deterimens the duty cycle of each
  // oversampled sample and therefore, it deterimines the volume of the tone.
  
  // count the number of wraps to deterimine the level
  static int nr_wraps_det_level = 0;

  // whether the level is going up or not
  static bool going_up = true;

  // The second counter keeps track of the number of wraps to actually set the
  // level.  This counter determines the frequency of the tone by setting the
  // level high or low.  When the value is high, it is based on the level we
  // deterimed in the first counter.
  
  // count the number of wraps to set the level
  static int nr_wraps_set_level = 0;

  // whether we want to have a high signal or low signal
  static bool high = true;

  // the level (the volume)
  static int level = 0;

  // Clear the interrupt flag that brought us here
  pwm_clear_irq(pwm_gpio_to_slice_num(PIN_BUZZER));

  // every NR_WRAPS_DET_LEVEL wrap, we compute a new level between 16 and 128
  if (nr_wraps_det_level == NR_WRAPS_DET_LEVEL) {
    nr_wraps_det_level = 0;
    if (going_up) {
      level++;
      if (level > HIGH_LEVEL) {
	level = HIGH_LEVEL;
	going_up = false;
      }
    }
    else {
      level--;
      if (level < LOW_LEVEL) {
	level = LOW_LEVEL;
	going_up = true;
      }
    }
  }
  else {
    nr_wraps_det_level++;
  }

  // every time we are at a wrap that allows us to set a new level or turn the
  // level to 0
  // Note that this frequency is independent of the period of dimming.
  if (nr_wraps_set_level == (SAMPLE_RATE * OVERSAMPLING_FACTOR / NOTE_HZ / 2)) {
    // set the number of samples to zero
    nr_wraps_set_level = 0;

    // set the level
    if (high) {
      pwm_set_gpio_level(PIN_BUZZER, level);
      high = false;
    }
    else {
      pwm_set_gpio_level(PIN_BUZZER, 0);
      high = true;
    }
  }
  else {
    nr_wraps_set_level++;
  }
}

int main() {
  stdio_init_all();

  // Tell the pin that the PWM is in charge of its value.
  gpio_set_function(PIN_BUZZER, GPIO_FUNC_PWM);

  // The slice number give the pin of the buzzer
  uint slice_num = pwm_gpio_to_slice_num(PIN_BUZZER);

  // Set an IRQ handler for every wrap of the counter
  pwm_clear_irq(slice_num);
  pwm_set_irq_enabled(slice_num, true);
  irq_set_exclusive_handler(PWM_IRQ_WRAP, on_pwm_wrap);
  irq_set_enabled(PWM_IRQ_WRAP, true);
  
  // Get some sensible defaults for the slice configuration. By default, the
  // counter is allowed to wrap over its maximum range (0 to 2**16-1)
  pwm_config config = pwm_get_default_config();

  // set the wrap value a little bit higher than 128, so we can use this number
  // and we will never reach a 100% duty cycle
  uint16_t wrap = WRAP;
  
  pwm_config_set_wrap(&config, wrap);
  pwm_init(slice_num, &config, true);

  // sleep 3 seconds to make sure we can see the output of the prints.
  sleep_ms(3000);

  //printf("system clock: %d\n", clock_get_hz(clk_sys));
  printf("system clock: %d\n", FREQ_CLOCK_SYS_BUS);
  printf("sample rate: %d\n", SAMPLE_RATE);
  printf("oversampling factor: %d\n", OVERSAMPLING_FACTOR);
  printf("wrap: %d\n", wrap);
  printf("time_wrap: %e\n", TIME_WRAP);
  printf("# wraps to determine a new level: %d\n", NR_WRAPS_DET_LEVEL);

  
  while (1) {
    tight_loop_contents();
  }
}
